package cn.com.qmth.examcloud.starters.face.verify.service.faceplus;

import cn.com.qmth.examcloud.starters.face.verify.FaceVerifyProperties;
import cn.com.qmth.examcloud.starters.face.verify.common.Constants;
import cn.com.qmth.examcloud.starters.face.verify.common.FaceResult;
import cn.com.qmth.examcloud.starters.face.verify.common.HttpClientBuilder;
import cn.com.qmth.examcloud.starters.face.verify.common.JsonHelper;
import okhttp3.FormBody;
import okhttp3.Request;
import okhttp3.Response;
import okhttp3.ResponseBody;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.Map;

public class FacePlusApiHelper {

    private final static Logger log = LoggerFactory.getLogger(FacePlusApiHelper.class);

    public static FaceResult faceDetect(FaceVerifyProperties properties, Map<String, String> params) {
        FormBody.Builder formBody = new FormBody.Builder();
        formBody.add("api_key", properties.getFacePlusKey());
        formBody.add("api_secret", properties.getFacePlusSecret());

        for (Map.Entry<String, String> entry : params.entrySet()) {
            formBody.add(entry.getKey(), entry.getValue());
        }

        Request.Builder request = new Request.Builder().url(Constants.FACEPP_FACE_DETECT_URL).post(formBody.build());
        FacePlusResponse response = callApi(Constants.FACEPP_FACE_DETECT_URL, request);

        FaceResult result = new FaceResult();
        result.setPass(false);
        result.setApiLimit(false);

        if (StringUtils.isNotEmpty(response.getError_message())) {
            result.setError(response.getError_message());

            // 请求并发是否超限(约20个/秒)
            if ("CONCURRENCY_LIMIT_EXCEEDED".equals(response.getError_message())) {
                result.setApiLimit(true);
            }
            return result;
        }

        int faceNum = response.getFace_num() != null ? response.getFace_num() : 0;
        result.setFaceNum(faceNum);

        if (faceNum == 1) {
            // 是否人脸检测通过 (只有一张人脸算成功，否则算失败)
            result.setPass(true);
        }

        return result;
    }

    public static FaceResult faceCompare(FaceVerifyProperties properties, Map<String, String> params) {
        FormBody.Builder formBody = new FormBody.Builder();
        formBody.add("api_key", properties.getFacePlusKey());
        formBody.add("api_secret", properties.getFacePlusSecret());

        for (Map.Entry<String, String> entry : params.entrySet()) {
            formBody.add(entry.getKey(), entry.getValue());
        }

        Request.Builder request = new Request.Builder().url(Constants.FACEPP_FACE_COMPARE_URL).post(formBody.build());
        FacePlusResponse response = callApi(Constants.FACEPP_FACE_COMPARE_URL, request);

        FaceResult result = new FaceResult();
        result.setPass(false);
        result.setApiLimit(false);

        if (StringUtils.isNotEmpty(response.getError_message())) {
            result.setError(response.getError_message());

            // 请求并发是否超限(约20个/秒)
            if ("CONCURRENCY_LIMIT_EXCEEDED".equals(response.getError_message())) {
                result.setApiLimit(true);
            }
            return result;
        }

        FacePlusThresholds thresholds = response.getThresholds();
        if (thresholds != null) {
            float confidence = response.getConfidence() != null ? response.getConfidence() : 0f;

            result.setScore((double) confidence);
            result.setThresholds(new JsonHelper().toJson(thresholds));

            // 是否人脸对比通过
            if (confidence >= thresholds.getLe_4()) {
                result.setPass(true);
            }
        }

        return result;
    }

    private static FacePlusResponse callApi(String requestUrl, Request.Builder request) {
        try (Response response = HttpClientBuilder.getClient().newCall(request.build()).execute();
             ResponseBody body = response.body();) {
            String bodyStr = body.string();

            // System.out.println(response.code() + " - " + bodyStr);

            if (response.isSuccessful() || response.code() == 401 || response.code() == 403) {
                FacePlusResponse result = new JsonHelper().parseJson(bodyStr, FacePlusResponse.class);
                if (result != null) {
                    return result;
                }
            }

            log.warn("[Face++] url:{} response:{} body:{}", requestUrl, response.code(), bodyStr);
        } catch (IOException e) {
            log.error(e.getMessage(), e);
        }

        throw new RuntimeException("[Face++] 接口调用异常！");
    }

}
